// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MMESH_MMESH_HH
#define DUNE_MMESH_MMESH_HH

/** \file
 * \brief The MMesh include
 */

#include "grid/declaration.hh"
#include "cgal/defaults.hh"
#include "cgal/triangulationwrapper.hh"
#include "grid/mmesh.hh"

#endif // DUNE_MMESH_MMESH_HH
